import polars as pl
import re


@pl.api.register_dataframe_namespace("name_ext")
class NameExtensionNameSpace:
    "Functions that extend the Name capabilities of polars DataFrames"

    def __init__(self, df: pl.DataFrame):
        self._df = df

    def to_pascal_case(self) -> pl.DataFrame:
        "Converts column names to PascalCase"
        def _to_pascal_case(name: str) -> str:
            return "".join(
                word.capitalize() for word in re.sub(r"[_\s]+", " ", name).split()
            )

        columns = self._df.columns
        new_columns = {col: _to_pascal_case(col) for col in columns}
        return self._df.rename(new_columns)

    def to_snake_case(self) -> pl.DataFrame:
        "Converts column names to snake_case"
        def _to_snake_case(name: str) -> str:
            return re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()

        new_columns = {col: _to_snake_case(col) for col in self._df.columns}
        return self._df.rename(new_columns)

    def to_camel_case(self) -> pl.DataFrame:
        "Converts column names to camelCase"
        def _to_camel_case(name: str) -> str:
            words = re.sub(r"[_\s]+", " ", name).split()
            return words[0].lower() + "".join(word.capitalize() for word in words[1:])

        new_columns = {col: _to_camel_case(col) for col in self._df.columns}
        return self._df.rename(new_columns)

    def to_pascal_snake_case(self) -> pl.DataFrame:
        "Converts column names to Pascal_Snake_Case"
        def _to_pascal_snake_case(name: str) -> str:
            words = re.sub(r"[_\s]+", " ", name).split()
            return "_".join(word.capitalize() for word in words)

        new_columns = {col: _to_pascal_snake_case(col) for col in self._df.columns}
        return self._df.rename(new_columns)

    def to_kebeb_case(self) -> pl.DataFrame:
        "Converts column names to kebab-case"
        def _to_kebeb_case(name: str) -> str:
            return re.sub(r"(?<!^)(?=[A-Z])", "-", name).lower().replace("_", "-")

        new_columns = {col: _to_kebeb_case(col) for col in self._df.columns}
        return self._df.rename(new_columns)

    def to_upper_snake_case(self) -> pl.DataFrame:
        "Converts column names to UPPER_SNAKE_CASE"
        def _to_upper_snake_case(name: str) -> str:
            return re.sub(r"(?<!^)(?=[A-Z])", "_", name).upper().replace("-", "_")

        new_columns = {col: _to_upper_snake_case(col) for col in self._df.columns}
        return self._df.rename(new_columns)

    def to_train_case(self) -> pl.DataFrame:
        "Converts column names to Train-Case"
        def _to_train_case(name: str) -> str:
            return "-".join(
                word.capitalize() for word in re.sub(r"[_\s]+", " ", name).split()
            )

        new_columns = {col: _to_train_case(col) for col in self._df.columns}
        return self._df.rename(new_columns)



import json
from typing import Union
import polars as pl

def write_schema(schema: Union[pl.DataFrame, pl.Schema], file: str):
    "Saves a Polars schema a JSON file"
    if isinstance(schema, pl.DataFrame):
        schema = schema.schema

    stringified_values = [str(value) for value in schema.dtypes()]
    schema_dict = dict(zip(schema.names(), stringified_values))

    with open(file, "w") as f:
        json.dump(schema_dict, f)
    return

def read_schema(file: str):
    "Opens a JSON Schema file and return a Polars Schema object"
    with open(file, "r") as f:
        schema = json.load(f)
    
    schema_dict = {}
    for k, v in schema.items():
        try:
            schema_dict[k] = getattr(pl, v)
        except AttributeError:
            raise ValueError(f"Invalid type {v} for column {k}")
    
    schema_object = pl.Schema(schema_dict)
    return schema_object


import polars as pl

@pl.api.register_dataframe_namespace("str_ext")
class StringExtensionNamespace:
    """String Extensions for the Polars Library"""

    def __init__(self, df: pl.DataFrame):
        self._df = df

    def f1_string_similarity(self, col_a: str, col_b: str) -> pl.DataFrame:
        """
        Calculates a similarity score between two columns of strings based on common characters,
        accounting for repeated characters.
        
        Parameters:
        col_a (str): The name of the first column to compare.
        col_b (str): The name of the second column to compare.
        
        Returns:
        pl.DataFrame: A DataFrame with the similarity scores as a new column.
        """

        def similarity(row_str_a: str, row_str_b: str) -> float:
            # Normalize both strings (case-insensitive comparison)
            row_str_a = row_str_a.lower()
            row_str_b = row_str_b.lower()

            # If strings are identical, return a score of 1.0
            if row_str_a == row_str_b:
                return 1.0

            list1 = list(row_str_a)
            list2 = list(row_str_b)

            list2_copy = list2[:]
            intersection = []

            # Account for repeated characters by checking all occurrences
            for char in list1:
                if char in list2_copy:
                    intersection.append(char)
                    list2_copy.remove(char)
            
            common_chars = len(intersection)
            total_chars = len(list1) + len(list2)
            return (2 * common_chars) / total_chars if total_chars > 0 else 0.0

        # Apply the similarity function row-by-row
        similarity_scores = [
            similarity(row_a, row_b) for row_a, row_b in zip(self._df[col_a], self._df[col_b])
        ]

        # Add the similarity scores as a new column to the DataFrame
        self._df = self._df.with_columns(
            pl.Series("f1_score", similarity_scores)
        )

        return self._df


import polars as pl
from shapely import wkb, wkt
from shapely.geometry.base import BaseGeometry

@pl.api.register_dataframe_namespace("geo")
class GeometryExtensionNamespace:
    """Geometry utilities for handling WKB, WKT, and coordinate conversion."""

    def __init__(self, df: pl.DataFrame):
        self._df = df

    def _geom_to_coords(self, geom: BaseGeometry):
        """Convert any shapely geometry to a nested coordinate list."""
        if geom.geom_type == "Point":
            return list(geom.coords[0])
        elif geom.geom_type in {"LineString", "LinearRing"}:
            return [list(coord) for coord in geom.coords]
        elif geom.geom_type == "Polygon":
            exterior = [list(coord) for coord in geom.exterior.coords]
            interiors = [[list(coord) for coord in ring.coords] for ring in geom.interiors]
            return [exterior] + interiors if interiors else [exterior]
        elif geom.geom_type.startswith("Multi") or geom.geom_type == "GeometryCollection":
            return [self._geom_to_coords(part) for part in geom.geoms]
        else:
            return None  # Unknown type

    def wkb_to_coords(self, col: str, output_col: str = "coords") -> pl.DataFrame:
        coords = [
            self._geom_to_coords(wkb.loads(bytes.fromhex(val))) if val else None
            for val in self._df[col]
        ]
        return self._df.with_columns(pl.Series(output_col, coords))

    def coords_to_wkb(self, col: str, output_col: str = "wkb") -> pl.DataFrame:
        from shapely.geometry import shape
        wkb_hex = [
            shape(geom).wkb.hex() if geom else None
            for geom in self._df[col]
        ]
        return self._df.with_columns(pl.Series(output_col, wkb_hex))

    def wkt_to_coords(self, col: str, output_col: str = "coords") -> pl.DataFrame:
        coords = [
            self._geom_to_coords(wkt.loads(val)) if val else None
            for val in self._df[col]
        ]
        return self._df.with_columns(pl.Series(output_col, coords))

    def coords_to_wkt(self, col: str, output_col: str = "wkt") -> pl.DataFrame:
        from shapely.geometry import shape
        wkt_strs = [
            shape(geom).wkt if geom else None
            for geom in self._df[col]
        ]
        return self._df.with_columns(pl.Series(output_col, wkt_strs))



import polars as pl


@pl.api.register_expr_namespace("num_ext") 
class NumericExtensionNamespace: 
    def __init__(self, expr: pl.Expr):
        self._expr = expr

    # Roman numeral mappings
    _roman_map = [
        ("M", 1000), ("CM", 900), ("D", 500), ("CD", 400),
        ("C", 100), ("XC", 90), ("L", 50), ("XL", 40),
        ("X", 10), ("IX", 9), ("V", 5), ("IV", 4), ("I", 1)
    ]

    def to_roman(self) -> pl.Expr:
        """
        Convert an integer to Roman numerals.
        """
        def convert_to_roman(value: int) -> str:
            if not (0 < value < 4000):
                raise ValueError("Number out of range (1-3999)")

            result = []
            for roman, num in self._roman_map:
                while value >= num:
                    result.append(roman)
                    value -= num
            return ''.join(result)

        # Use map_elements for element-wise operation
        return self._expr.map_elements(convert_to_roman,return_dtype=pl.String)

    def from_roman(self) -> pl.Expr:
        """
        Convert Roman numerals to integers.
        """
        roman_to_value = {roman: value for roman, value in self._roman_map}

        def convert_from_roman(roman: str) -> int:
            i = 0
            total = 0
            while i < len(roman):
                # Check for two-character numeral first
                if i + 1 < len(roman) and roman[i:i+2] in roman_to_value:
                    total += roman_to_value[roman[i:i+2]]
                    i += 2
                else:
                    total += roman_to_value[roman[i]]
                    i += 1
            return total

        # Use map_elements for element-wise operation
        return self._expr.map_elements(convert_from_roman,return_dtype=pl.Int64)


from .io import *
from .name import *
from .numeric import *
from .string import *

__all__ = [
    "NameExtensionNameSpace",
    "NumericExtensionNamespace",
    "StringExtensionNamespace",
    "write_schema",
    "read_schema",
]
